

from typing import Any
from typing import List
from typing import Union
from typing import Optional
from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy import Integer
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy_json import mutable_json_type


from tendril.utils.db import DeclBase
from tendril.utils.db import BaseMixin
from tendril.utils.db import TimestampMixin
from tendril.utils.pydantic import TendrilTBaseModel
from .content_formats import MediaContentFormatModel
from .content_formats import ThumbnailListingTModel
from .content_formats import MediaContentFormatInfoTModel
from .content_formats import MediaContentFormatInfoFullTModel

from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class MediaContentInfoTModel(TendrilTBaseModel):
    # TODO Split this into a union of types
    content_type: str
    bg_color: Optional[Any]
    path: Optional[str]
    args: Optional[dict]
    formats: Optional[List[Union[MediaContentFormatInfoFullTModel,
                                 MediaContentFormatInfoTModel]]]
    thumbnails: Optional[ThumbnailListingTModel]
    default_duration: Optional[int]
    contents: Optional[List['SequenceMemberTModel']]


class MediaContentInfoFullTModel(MediaContentInfoTModel):
    estimated_duration: Any
    published: bool


class SequenceMemberTModel(TendrilTBaseModel):
    position: int
    duration: Optional[int]
    content: Union[MediaContentInfoFullTModel,
                   MediaContentInfoTModel]


MediaContentInfoTModel.update_forward_refs()
MediaContentInfoFullTModel.update_forward_refs()


class ContentModel(DeclBase, BaseMixin, TimestampMixin):
    type_name = 'generic'
    type_description = "Other"
    content_type = Column(String(32), nullable=False, default=type_name)
    allows_actual_media = False

    bg_color = Column(String(20))

    # TODO device_content and advertisement are instance specific.
    #  These need to be moved into sxm-core somehow.
    @declared_attr
    def device_content(cls):
        return relationship("DeviceContentModel", uselist=False, back_populates='content', lazy='select')

    @declared_attr
    def advertisement(cls):
        return relationship("AdvertisementModel", uselist=False, back_populates='content', lazy='select')

    @property
    def interest(self):
        return self.advertisement or self.device_content

    sequence_usages: Mapped[List["SequenceContentAssociationModel"]] = relationship()

    __mapper_args__ = {
        "polymorphic_identity": type_name,
        "polymorphic_on": content_type
    }

    def export(self, full=False, explicit_durations_only=False):
        rv = {'content_type': self.content_type,
              'estimated_duration': self.estimated_duration()}
        if self.bg_color:
            rv['bg_color'] = self.bg_color
        return rv

    def is_usable(self):
        return False

    def estimated_duration(self):
        return None


class MediaContentModel(ContentModel):
    type_name = 'media'
    type_description = "Single Media File"
    id = Column(Integer, ForeignKey("Content.id"), primary_key=True)
    allows_actual_media = True
    fidx = Column(Integer, default=0, nullable=False)

    @declared_attr
    def formats(cls):
        return relationship(MediaContentFormatModel, back_populates="content", lazy="selectin")

    __mapper_args__ = {
        "polymorphic_identity": type_name
    }

    def export(self, full=False, explicit_durations_only=False):
        rv = super(MediaContentModel, self).export(full=full, explicit_durations_only=explicit_durations_only)
        rv['formats'] = [x.export(full=full) for x in self.formats]
        for fmt in self.formats:
            if len(fmt.thumbnails):
                rv['thumbnails'] = fmt.export_thumbnails()
                break
        return rv

    def estimated_duration(self):
        durations = [x.duration for x in self.formats]
        simple_durations = [x for x in durations if x > 0]
        step_durations = [x for x in durations if x < 0]
        if not len(simple_durations) and not len(step_durations):
            return 0
        if not len(step_durations):
            return max(simple_durations)
        if not len(simple_durations):
            return min(step_durations)
        step_durations = [x * -1 * 10 for x in step_durations]
        return max(simple_durations + step_durations)

    def is_usable(self):
        return len(self.formats) > 0


class StructuredContentModel(ContentModel):
    type_name = 'structured'
    type_description = "Device Assembled Structured Content"

    id = Column(Integer, ForeignKey("Content.id"), primary_key=True)
    path = Column(String(64))
    args = Column(mutable_json_type(dbtype=JSONB))

    __mapper_args__ = {
        "polymorphic_identity": type_name
    }

    def export(self, full=False, explicit_durations_only=False):
        rv = super(StructuredContentModel, self).export(full=full, explicit_durations_only=explicit_durations_only)
        rv['path'] = self.path
        if self.args:
            rv['args'] = self.args
        return rv

    def estimated_duration(self):
        return None

    def is_usable(self):
        # TODO Check if the path corresponds to a valid provider as well?
        return self.path is not None


class SequenceContentModel(ContentModel):
    type_name = 'sequence'
    type_description = "Sequence of other content types"

    id = Column(Integer, ForeignKey("Content.id"), primary_key=True)
    default_duration = Column(Integer, nullable=False, default=10)

    contents: Mapped[List["SequenceContentAssociationModel"]] = \
        relationship(order_by="SequenceContentAssociationModel.position")

    __mapper_args__ = {
        "polymorphic_identity": type_name
    }

    def export(self, full=False, explicit_durations_only=False):
        rv = super(SequenceContentModel, self).export(full=full, explicit_durations_only=explicit_durations_only)
        rv['default_duration'] = self.default_duration
        rv['contents'] = [x.export(full=full, explicit_durations_only=explicit_durations_only) for x in self.contents]
        return rv

    def estimated_duration(self):
        durations = [x.duration or x.content.estimated_duration() for x in self.contents]
        actual_durations = []
        for duration in durations:
            if duration < 0:
                duration = self.default_duration * -1 * duration
            actual_durations.append(duration)
        return sum(actual_durations) + len(durations)

    def is_usable(self):
        return len(self.contents) > 0 and all([x.content.is_usable() for x in self.contents])


class SequenceContentAssociationModel(DeclBase):
    __tablename__ = "SequenceContentAssociation"
    sequence_id: Mapped[int] = mapped_column(ForeignKey("SequenceContent.id"), primary_key=True)
    content_id: Mapped[int] = mapped_column(ForeignKey("Content.id"))
    position: Mapped[int] = mapped_column(primary_key=True)
    duration: Mapped[Optional[int]]
    sequence: Mapped[SequenceContentModel] = relationship(back_populates="contents", foreign_keys=[sequence_id], lazy='selectin')
    content: Mapped[ContentModel] = relationship(back_populates="sequence_usages", foreign_keys=[content_id], lazy='joined')

    def export(self, full=False, explicit_durations_only=False):
        return {
            'position': self.position,
            'duration': self.duration,
            'content': self.content.export(full=full)
        }





from sqlalchemy.exc import NoResultFound
from tendril.db.controllers.interests import get_interest

from tendril.utils.db import with_db
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class ContentProviderBase(object):
    name = 'base'
    display_name = 'Base Content Provider'
    interest_class = None
    path = None
    args = {}
    requires_app = None

    def __init__(self):
        self._interest = None

    @with_db
    def commit_to_db(self, session=None):
        try:
            self._interest = get_interest(name=self.name,
                                          type=self.interest_class,
                                          session=session)
            # TODO Implement update here.
            #  Right now these things are purely write-only.
        except NoResultFound:
            self._interest = self.interest_class(name=self.name,
                                                 path=self.path, iargs=self.args,
                                                 requires_app=self.requires_app,
                                                 must_create=True, session=session)
            self._interest.set_descriptive_name(self.display_name, session=session)
            self._auto_activate(session=session)

    @with_db
    def _auto_activate(self, session=None):
        session.flush()
        from tendril.interests import Platform
        p = Platform(get_interest(id=1, session=session))
        p.add_child(self._interest.model_instance, auth_user=1, session=session)
        self._interest.activate(auth_user=1, session=session)


def load(manager):
    pass





import importlib

from .base import ContentProviderBase

from tendril.utils.versions import get_namespace_package_names
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class ContentProviderManager(object):
    def __init__(self, prefix):
        self._prefix = prefix
        self._providers = {}
        self._register_providers()
        self.finalized = False

    def _register_providers(self):
        logger.debug("Loading content providers from {0}".format(self._prefix))
        modules = list(get_namespace_package_names(self._prefix))
        for m_name in modules:
            if m_name == __name__:
                continue
            m = importlib.import_module(m_name)
            logger.debug("Loading content providers from {0}".format(m_name))
            m.load(self)

    def register_provider(self, provider : ContentProviderBase):
        self._providers[provider.name] = provider

    def finalize(self):
        for name, provider in self._providers.items():
            logger.info(f"Installing Content Provider '{name}'")
            provider.commit_to_db()
        self.finalized = True

    @property
    def registered_providers(self):
        return self._providers





from tendril.common.exceptions import HTTPCodedException
from tendril.common.interests.exceptions import InterestActionException
from tendril.common.interests.exceptions import InterestTypeUnsupported


class ContentNotReady(InterestActionException):
    status_code = 406

    def __init__(self, *args, **kwargs):
        super(ContentNotReady, self).__init__(*args, **kwargs)

    def __str__(self):
        return f"The interest {self.interest_id}, {self.interest_name} does not have " \
               f"an associated content containter.'{self.action}' cannot be performed."


class ContentTypeMismatchError(InterestActionException):
    status_code = 406

    def __init__(self, content_type, required, *args, **kwargs):
        super(ContentTypeMismatchError, self).__init__(*args, **kwargs)
        self.content_type = content_type
        self.required = required

    def __str__(self):
        return f"The interest {self.interest_id}, {self.interest_name} is defined as " \
               f"as content_type '{self.content_type}'. The action `{self.action}` is " \
               f"unsupported. Requires `{self.required}`."


class FileTypeUnsupported(InterestActionException):
    status_code = 406

    def __init__(self, extension, allowed, *args, **kwargs):
        super(FileTypeUnsupported, self).__init__(*args, **kwargs)
        self.extension = extension
        self.allowed = allowed

    def __str__(self):
        return f"Media format upload to interest {self.interest_id}, {self.interest_name} " \
               f"failed. Provided file has extension '{self.extension}' which is unsupported. " \
               f"Supported extensions are `{self.allowed}`."






from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import declared_attr
from sqlalchemy.orm import relationship

from tendril.filestore.db.model import StoredFileModel
from tendril.utils.db import DeclBase
from tendril.utils.db import BaseMixin
from tendril.utils.db import TimestampMixin

from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class MediaContentFormatThumbnailModel(DeclBase, BaseMixin, TimestampMixin):
    stored_file_id: Mapped[int] = mapped_column(ForeignKey("StoredFile.id"), nullable=False)
    format_id: Mapped[int] = mapped_column(ForeignKey('MediaContentFormat.id'), nullable=False)
    width = Column(Integer, nullable=False)
    height = Column(Integer, nullable=False)

    @declared_attr
    def format(cls):
        return relationship('MediaContentFormatModel', back_populates="thumbnails", lazy="selectin")

    def export(self, full=False):
        return {f'{self.width}x{self.height}': self.stored_file.expose_uri}

    @declared_attr
    def stored_file(cls):
        return relationship(StoredFileModel, lazy="selectin")





from typing import Any
from typing import Dict
from typing import Optional
from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy import Integer
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import declared_attr
from sqlalchemy.orm import relationship
from sqlalchemy_json import mutable_json_type
from sqlalchemy.dialects.postgresql import JSONB

from tendril.filestore.db.model import StoredFileModel
from tendril.utils.db import DeclBase
from tendril.utils.db import BaseMixin
from tendril.utils.db import TimestampMixin
from tendril.utils.pydantic import TendrilTBaseModel

from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class StoredFileHashTModel(TendrilTBaseModel):
    sha256: Optional[str]

class MediaContentFormatInfoTModel(TendrilTBaseModel):
    format_class: str
    format_id: int
    width: Optional[int]
    height: Optional[int]
    duration: Optional[int]
    uri: str
    hash: StoredFileHashTModel
    published: Optional[bool]


ThumbnailListingTModel = Dict[str, str]


class MediaContentFormatInfoFullTModel(MediaContentFormatInfoTModel):
    info: Any
    thumbnails: ThumbnailListingTModel


class MediaContentFormatModel(DeclBase, BaseMixin, TimestampMixin):
    format_class_name = 'generic'
    format_class = Column(String(32), nullable=False)

    content_id: Mapped[int] = mapped_column(ForeignKey('MediaContent.id'))
    width = Column(Integer, nullable=True)
    height = Column(Integer, nullable=True)
    duration = Column(Integer, nullable=True)
    info = Column(mutable_json_type(dbtype=JSONB, nested=True), default={})

    @declared_attr
    def content(cls):
        return relationship('MediaContentModel',
                            back_populates="formats", lazy="selectin")

    @declared_attr
    def thumbnails(cls):
        return relationship('MediaContentFormatThumbnailModel',
                            back_populates='format', lazy='selectin')

    def export_thumbnails(self):
        rv = {}
        for t in self.thumbnails:
            rv.update(t.export())
        return rv

    def export(self, full=False):
        rv = {'format_class': self.format_class_name,
              'format_id': self.id,
              'duration': self.duration}
        if self.width or self.height:
            rv['width'] = self.width
            rv['height'] = self.height
            if full:
                rv['info'] = self.info
        if full:
            rv['thumbnails'] = self.export_thumbnails()
        return rv

    def estimated_duration(self):
        if self.duration:
            if self.duration < 0:
                return self.duration * -10
            return self.duration
        else:
            return 10

    __mapper_args__ = {
        "polymorphic_identity": format_class_name,
        "polymorphic_on": format_class
    }


class ExternalPublishedMediaContentFormatModel(MediaContentFormatModel):
    format_class_name = 'external_published'
    id = Column(Integer, ForeignKey("MediaContentFormat.id"), primary_key=True)
    uri = Column(String, nullable=False)

    def export(self, full=False):
        rv = super(ExternalPublishedMediaContentFormatModel, self).export(full=full)
        rv['uri'] = self.uri
        return rv

    __mapper_args__ = {
        "polymorphic_identity": format_class_name,
    }


class FileMediaContentFormatModel(MediaContentFormatModel):
    format_class_name = 'file_media'
    id = Column(Integer, ForeignKey("MediaContentFormat.id"), primary_key=True)
    stored_file_id: Mapped[int] = mapped_column(ForeignKey("StoredFile.id"), nullable=False)

    @declared_attr
    def stored_file(cls):
        return relationship(StoredFileModel, lazy="selectin")

    def export(self, full=False):
        rv = super(FileMediaContentFormatModel, self).export(full=full)
        rv['uri'] = self.stored_file.expose_uri
        rv['hash'] = self.stored_file.fileinfo['hash']
        return rv

    __mapper_args__ = {
        "polymorphic_identity": format_class_name,
    }





import os
import asyncio
from asgiref.sync import async_to_sync

from httpx import HTTPStatusError

from tendril.db.controllers.interests import get_interest
from tendril.common.content.exceptions import ContentTypeMismatchError

from tendril.filestore import buckets
from tendril.config import MEDIA_UPLOAD_FILESTORE_BUCKET
from tendril.config import MEDIA_PUBLISHING_FILESTORE_BUCKET

from tendril.interests.base import InterestBase
from tendril.common.states import LifecycleStatus
from tendril.authz.roles.interests import require_state
from tendril.authz.roles.interests import require_permission

from tendril.caching import tokens
from tendril.caching.tokens import TokenStatus

from tendril.structures.content import content_types
from tendril.db.models.content import ContentModel
from tendril.db.controllers.content import create_content
from tendril.db.controllers.content import create_content_format_file
from tendril.db.controllers.content import create_content_format_thumbnail
from tendril.db.controllers.content import sequence_next_position
from tendril.db.controllers.content import sequence_heal_positions
from tendril.db.controllers.content import sequence_get_contents
from tendril.db.controllers.content import sequence_add_content
from tendril.db.controllers.content import sequence_remove_content
from tendril.common.content.exceptions import ContentNotReady
from tendril.common.interests.representations import rewrap_interest
from tendril.common.interests.representations import ExportLevel

from tendril.utils.parsers.media.info import get_media_info
from tendril.utils.parsers.media.thumbnails import generate_thumbnails

from tendril.utils.fsutils import TEMPDIR
from tendril.utils.db import with_db
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class MediaContentInterest(InterestBase):
    # WARNING.
    #
    # This is a common container for the different content types.
    # This is terrible, and needs to be fixed. The main issue is
    # the current implicit assumption of a 1-1 mapping between an
    # interest class and the model.
    #
    # Devices should have the same problem, but more effort has
    # gone into working around it there, I think.
    #
    # The resolution will probably be to extend the same kind of
    # class composition that was done at the interest level to the
    # interest class as well.

    token_namespace = 'mfu'
    upload_bucket_name = MEDIA_UPLOAD_FILESTORE_BUCKET
    publish_bucket_name = MEDIA_PUBLISHING_FILESTORE_BUCKET
    additional_creation_fields = ['content_type']
    additional_export_fields = ['content_type']
    additional_activation_checks = ['check_content_usable']

    def __init__(self, *args, content_type=None, **kwargs):
        self._content_type = content_type
        super(MediaContentInterest, self).__init__(*args, **kwargs)
        self._upload_bucket = None
        self._publish_bucket = None

    # Common
    @property
    def content_type(self):
        if not self._content_type:
            self._content_type = self._model_instance.content.content_type
        return self._content_type

    @with_db
    def set_content_type(self, content_type, session=None):
        if self._content_type and self._content_type != content_type:
            raise ValueError(f"Content Type {self._content_type} already set for interest {self.id} "
                             f"and cannot be changed. Create a new device_content instead.")
        if content_type not in content_types:
            raise ValueError(f"Content Type {content_type} is not recognized. Try one of {content_types}")
        self._content_type = content_type
        self._commit_to_db(session=session)

    @property
    def content(self):
        return self.model_instance.content

    @with_db
    def check_content_usable(self, auth_user=None, session=None):
        return self.content.is_usable()

    @with_db
    def activate(self, background_tasks=None, auth_user=None, session=None):
        result, msg = super().activate(background_tasks=background_tasks,
                                       auth_user=auth_user, session=session)

        if not self.model_instance.status == LifecycleStatus.ACTIVE:
            return result, msg

        publishable = self.publishable()

        if background_tasks:
            background_tasks.add_task(self._publish_files, publishable)
        else:
            asyncio.ensure_future(self._publish_files(publishable))
        return result, msg

    async def _publish_files(self, stored_files):
        for stored_file in stored_files:
            logger.info(f"Publishing file {stored_file.filename}")
            try:
                upload_response = await self.upload_bucket.move(
                    filename=stored_file.filename,
                    target_bucket=self.publish_bucket_name,
                    actual_user=None,
                )
            except HTTPStatusError as e:
                self._report_filestore_error(None, e, "Publishing media file")
                continue

    def publishable(self):
        content = self.model_instance.content
        rv = []
        if hasattr(content, "formats"):
            for fmt in self.model_instance.content.formats:
                if hasattr(fmt, 'stored_file'):
                    if fmt.stored_file.bucket.name == self.upload_bucket_name:
                        rv.append(fmt.stored_file)
                for thumbnail in fmt.thumbnails:
                    if thumbnail.stored_file.bucket.name == self.upload_bucket_name:
                        rv.append(thumbnail.stored_file)
        return rv

    def published(self):
        if self.status != LifecycleStatus.ACTIVE:
            return False
        content = self.model_instance.content
        if hasattr(content, "formats"):
            for fmt in self.model_instance.content.formats:
                if hasattr(fmt, 'stored_file'):
                    if fmt.stored_file.bucket.name != self.publish_bucket_name:
                        return False
        return True

    @with_db
    @require_state((LifecycleStatus.ACTIVE, LifecycleStatus.APPROVAL, LifecycleStatus.NEW))
    @require_permission('read_artefacts', strip_auth=False, required=False)
    def content_information(self, full=False, auth_user=None, session=None):
        if not self._model_instance.content:
            raise ContentNotReady('read_content_info', self.id, self.name)
        else:
            content: ContentModel = self._model_instance.content
            rv = content.export(full=full)
            if full:
                if 'formats' in rv.keys():
                    for fmt_info in rv['formats']:
                        fmt_info['published'] = self.format_published(fmt_info['format_id'])
                rv['published'] = self.published()
            return rv

    @with_db
    @require_state((LifecycleStatus.ACTIVE, LifecycleStatus.APPROVAL, LifecycleStatus.NEW))
    @require_permission('read_artefacts', strip_auth=False)
    def estimated_duration(self, auth_user=None, session=None):
        return self._model_instance.content.estimated_duration()

    @with_db
    def _commit_to_db(self, must_create=False, can_create=True, session=None):
        super(MediaContentInterest, self)._commit_to_db(must_create=must_create,
                                                        can_create=can_create,
                                                        session=session)
        if not self.model_instance.content_id:
            content = create_content(type=self._content_type, session=session)
            self.model_instance.content_id = content.id
            session.add(self.model_instance)
            session.commit()
            session.flush()

    # Media Content

    @property
    def upload_bucket(self):
        if not self._upload_bucket:
            self._upload_bucket = buckets.get_bucket(self.upload_bucket_name)
        return self._upload_bucket

    @property
    def publish_bucket(self):
        if not self._publish_bucket:
            self._publish_bucket = buckets.get_bucket(self.publish_bucket_name)
        return self._publish_bucket

    @property
    def formats(self):
        return self.content.formats

    @with_db
    @require_state((LifecycleStatus.NEW))
    @require_permission('add_artefact')
    def fidx_burn(self, session=None):
        rv = self.content.fidx
        self.content.fidx += 1
        session.add(self.content)
        return rv

    def _report_filestore_error(self, token_id, e, action_comment):
        logger.warn(f"Exception while {action_comment} : HTTP {e.response.status_code} {e.response.text}")
        if token_id:
            tokens.update(
                self.token_namespace, token_id, state=TokenStatus.FAILED,
                error={"summary": f"Exception while {action_comment}",
                       "filestore": {
                           "code": e.response.status_code,
                           "content": e.response.json()}
                       }
            )

    @with_db
    @require_state((LifecycleStatus.NEW))
    @require_permission('add_artefact', strip_auth=False)
    def add_format(self, file, rename_to=None, token_id=None, auth_user=None, session=None):
        storage_folder = f'{self.id}'
        if token_id:
            tokens.update(self.token_namespace, token_id,
                          state=TokenStatus.INPROGRESS, max=6,
                          current="Parsing Media Information")

        # 1. Parse Media Information
        filename = rename_to or file.filename
        media_info = get_media_info(file.file, filename=filename, original_filename=file.filename)

        if token_id:
            tokens.update(self.token_namespace, token_id,
                          current="Uploading Media File to Filestore", done=1)

        # 2. Upload File to Bucket
        try:
            upload_response = async_to_sync(self.upload_bucket.upload)(
                file=(os.path.join(storage_folder, filename), file.file),
                actual_user=auth_user.id, interest=self.id
            )
        except HTTPStatusError as e:
            self._report_filestore_error(token_id, e, "uploading media file to bucket")
            return

        if token_id:
            tokens.update(self.token_namespace, token_id,
                          current="Generating Thumbnails", done=2)

        # 3. Generate Thumbnails

        thumbnail_folder = os.path.join(TEMPDIR, os.path.splitext(filename)[0])
        os.makedirs(thumbnail_folder, exist_ok=True)
        generated_thumbnails = generate_thumbnails(file.file, thumbnail_folder, filename=filename)

        if token_id:
            tokens.update(self.token_namespace, token_id,
                          current="Uploading Thumbnails to Filestore", done=3)

        # 4. Upload Thumbnails to Bucket

        published_thumbnails = []
        for tsize, fpath in generated_thumbnails:
            fname = os.path.split(fpath)[1]
            with open(fpath, 'rb') as thumb_file:
                try:
                    response = async_to_sync(self.upload_bucket.upload)(
                        file=(os.path.join(storage_folder, fname), thumb_file),
                        actual_user=auth_user.id, interest=self.id
                    )
                except HTTPStatusError as e:
                    self._report_filestore_error(token_id, e, "uploading thumbnail to bucket")
                    return
            published_thumbnails.append((tsize, fname, response))

        if token_id:
            tokens.update(self.token_namespace, token_id,
                          current="Registering Media Format", done=4)

        # 5. Create Format DB Entry

        format_model_instance = create_content_format_file(
            id=self.model_instance.content_id,
            stored_file_id=upload_response['storedfileid'],
            width=media_info.width(),
            height=media_info.height(),
            duration=media_info.duration(),
            info=media_info.asdict(),
        )

        if token_id:
            tokens.update(self.token_namespace, token_id,
                          current="Registering Media Format Thumbnails", done=5,
                          metadata={'format_id': format_model_instance.id})

        # 6. Create Thumbnail DB Entries

        for tsize, fname, response in published_thumbnails:
            create_content_format_thumbnail(
                id=format_model_instance.id,
                stored_file_id=response['storedfileid'],
                width=tsize[0], height=tsize[1],
            )

        if token_id:
            tokens.update(self.token_namespace, token_id, current="Finishing", done=6)

        # 7. Close Upload Ticket
        tokens.close(self.token_namespace, token_id)

    def get_format(self, format_id):
        formats = self.model_instance.content.formats
        for candidate in formats:
            if candidate.id == format_id:
                return candidate

    def format_published(self, format_id):
        fmt = self.get_format(format_id)
        if hasattr(fmt, 'stored_file'):
            if fmt.stored_file.bucket.name != self.publish_bucket_name:
                return False
        return True

    @with_db
    @require_state((LifecycleStatus.ACTIVE, LifecycleStatus.APPROVAL, LifecycleStatus.NEW))
    @require_permission('read_artefacts', strip_auth=False)
    def format_information(self, format_id, full=False, auth_user=None, session=None):
        fmt = self.get_format(format_id)
        rv = fmt.export(full=full)
        if full:
            rv['published'] = self.format_published(format_id)
        return rv

    @with_db
    @require_state((LifecycleStatus.NEW))
    @require_permission('delete_artefact', strip_auth=False)
    def delete_format(self, format_id, auth_user=None, session=None):
        pass

    # Content Providers

    @with_db
    @require_state((LifecycleStatus.NEW))
    @require_permission('add_artefact', strip_auth=False)
    def generate_from_provider(self, provider_id, args, auth_user=None, session=None):
        if self.content_type != 'structured':
            raise ContentTypeMismatchError(self.content_type, 'structured',
                                           'add_artefact', self.id, self.name)

        provider = get_interest(id=provider_id, type='content_provider', session=session).actual
        generated = provider.generate(args, auth_user=auth_user, session=session)
        for k, v in generated.items():
            setattr(self.model_instance.content, k, v)
        session.add(self.model_instance.content)
        session.flush()
        return self.model_instance.content

    # Content Sequence

    @with_db
    @require_state((LifecycleStatus.NEW, LifecycleStatus.APPROVAL, LifecycleStatus.ACTIVE))
    @require_permission('add_artefact', strip_auth=False)
    def sequence_set_default_duration(self, default_duration=10, auth_user=None, session=None):
        if self.content_type != 'sequence':
            raise ContentTypeMismatchError(self.content_type, 'sequence',
                                           'add_artefact', self.id, self.name)

        if not isinstance(default_duration, int) or default_duration < 0:
            raise ValueError("Expecting a non-negative integer for duration")

        self.model_instance.content.default_duration = default_duration
        session.add(self.model_instance.content)
        session.flush()
        return {'interest_id': self.id,
                'default_duration': self.model_instance.content.default_duration}

    @with_db
    @require_state((LifecycleStatus.NEW, LifecycleStatus.APPROVAL, LifecycleStatus.ACTIVE))
    @require_permission('read', strip_auth=False)
    def sequence_get_contents(self, full=False, auth_user=None, session=None):
        if self.content_type != 'sequence':
            raise ContentTypeMismatchError(self.content_type, 'sequence',
                                           'read', self.id, self.name)

        contents = sequence_get_contents(id=self.model_instance.content_id,
                                         session=session)
        export_level = ExportLevel.STUB
        contents = [{'interest':
            rewrap_interest(x["content"].interest).export(
                export_level=export_level, auth_user=auth_user, session=session),
            'content_info': x["content"].export(explicit_durations_only=True),
        }
            for x in contents]

        return {'interest_id': self.id,
                'default_duration': self.model_instance.content.default_duration,
                'estimated_duration': self.estimated_duration(auth_user=auth_user,
                                                              session=session),
                'contents': contents}

    @with_db
    @require_state((LifecycleStatus.NEW))
    @require_permission('add_artefact', strip_auth=False)
    def sequence_add(self, content_id, position=None, duration=None, auth_user=None, session=None):
        if self.content_type != 'sequence':
            raise ContentTypeMismatchError(self.content_type, 'sequence',
                                           'add_artefact', self.id, self.name)

        # Get Content and Verify Access
        content = get_interest(content_id, type=self.type_name, session=session).actual
        if not content.check_user_access(auth_user, 'read', session=session):
            raise PermissionError("User does not seem to have access to the underlying content. "
                                  "Cannot add to sequence.")

        if not content.status == LifecycleStatus.ACTIVE:
            raise ValueError("The content must be active before it can be added to a sequence.")

        if not duration:
            _duration = content.estimated_duration(auth_user=auth_user, session=session)

        if not _duration:
            raise ValueError("We need a duration, however none is provided and the content does "
                             "not provide it intrinsically.")

        # Create and commit Association Model
        sequence_add_content(id=self.model_instance.content_id,
                             content=content.model_instance.content_id,
                             position=position,
                             duration=duration,
                             session=session)

        sequence_heal_positions(id=self.model_instance.content_id, session=session)
        return True

    @with_db
    @require_state((LifecycleStatus.NEW))
    @require_permission('add_artefact', strip_auth=False)
    def sequence_remove(self, position=None, auth_user=None, session=None):
        if self.content_type != 'sequence':
            raise ContentTypeMismatchError(self.content_type, 'sequence',
                                           'add_artefact', self.id, self.name)

        sequence_remove_content(id=self.model_instance.content_id,
                                position=position,
                                session=session)
        sequence_heal_positions(id=self.model_instance.content_id, session=session)
        return True





import os
from typing import Dict
from typing import Union
from typing import Optional
from pydantic.fields import Field
from inflection import singularize
from inflection import titleize

from fastapi import APIRouter
from fastapi import Request
from fastapi import Depends
from fastapi import File
from fastapi import Body
from fastapi import UploadFile
from fastapi import BackgroundTasks

from tendril.authn.users import auth_spec
from tendril.authn.users import AuthUserModel
from tendril.authn.users import authn_dependency

from tendril.apiserver.templates.base import ApiRouterGenerator
from tendril.utils.pydantic import TendrilTORMModel
from tendril.utils.pydantic import TendrilTBaseModel
from tendril.utils.db import get_session

from tendril.caching import tokens
from tendril.caching.tokens import GenericTokenTModel

from tendril.structures.content import content_models
from tendril.config import MEDIA_EXTENSIONS
from tendril.interests.mixins.content import MediaContentInterest
from tendril.common.content.exceptions import ContentTypeMismatchError
from tendril.common.content.exceptions import FileTypeUnsupported
from tendril.db.models.content_formats import MediaContentFormatInfoTModel
from tendril.db.models.content_formats import MediaContentFormatInfoFullTModel
from tendril.db.models.content import MediaContentInfoTModel
from tendril.db.models.content import MediaContentInfoFullTModel


class ContentTypeDetailTModel(TendrilTORMModel):
    type_description: str = Field(..., alias='display_name')


class SequenceDefaultDurationResponseTModel(TendrilTBaseModel):
    interest_id: int
    default_duration: int


class SequenceAddTModel(TendrilTBaseModel):
    content_id: int
    position: Optional[int]
    duration: Optional[int]


class InterestContentRouterGenerator(ApiRouterGenerator):
    def __init__(self, actual):
        super(InterestContentRouterGenerator, self).__init__()
        self._actual = actual

    async def accepted_types(self, request: Request,
                             user: AuthUserModel = auth_spec()):
        return self._actual.accepted_types

    async def content_info(self, request: Request,
                           id: int, full: bool = False,
                           user: AuthUserModel = auth_spec()):
        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)
            return interest.content_information(full=full, auth_user=user, session=session)

    async def upload_media_format(self, request: Request,
                                  id: int, background_tasks: BackgroundTasks,
                                  file: UploadFile = File(...),
                                  user: AuthUserModel = auth_spec()):
        """
        Warning : This can only be done when the interest is in the NEW state. This
                  enforces approval requirements on any change in the formats. An additional
                  API endpoint (something like reset approvals?) is needed for this.
        """
        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)

            # Make sure it's the correct content type before doing anything.
            if not content_models[interest.content_type].allows_actual_media:
                raise ContentTypeMismatchError(interest.content_type, 'media',
                                               'add_artefact', interest.id, interest.name,)

            # Ensure we accept the file extension
            file_ext = os.path.splitext(file.filename)[1]
            if file_ext not in MEDIA_EXTENSIONS:
                raise FileTypeUnsupported(file_ext, MEDIA_EXTENSIONS,
                                          'add_artefact', interest.id, interest.name,)

            # Get Auth clearance before sending the task to the background. This will
            # raise an exception if there is a problem.
            interest.add_format(probe_only=True, auth_user=user, session=session)

            # Burn an fidx and lock in the filename for the uploaded file.
            fidx = interest.fidx_burn(auth_user=user, session=session)
            storage_filename = f"{interest.name}_f{fidx}{file_ext}"

            # The above prechecks are required at the API level here since we are delegating
            # to a background task, and we want to avoid forcing the client to deal with
            # exceptions in that context.

            # TODO Consider providing a helper function in the interest iteself to do
            #  this stuff instead. The interest otherwise remains bare and unprotected.

            # Generate Upload Ticket and return
            upload_token = tokens.open(
                namespace='mfu',
                metadata={'interest_id': interest.id,
                          'filename': storage_filename},
                user=user.id, current="Request Created",
                progress_max=1, ttl=600,
            )

            background_tasks.add_task(interest.add_format, file=file,
                                      rename_to=storage_filename,
                                      token_id=upload_token.id,
                                      auth_user=user, session=session)

        return upload_token

    async def format_info(self, request: Request, id: int, format_id: int,
                          full: bool = True,
                          user: AuthUserModel = auth_spec()):
        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)
            return interest.format_information(format_id, full=full, auth_user=user, session=session)

    async def delete_media_format(self, request: Request,
                                  id: int, filename: str,
                                  user: AuthUserModel = auth_spec()):
        """
        Warning : This can only be done when the interest is in the NEW state. This
                  enforces approval requirements on any change in the formats. An additional
                  API endpoint (something like reset approvals?) is needed for this.
        """
        pass

    async def generate_provider_content(self, request:Request, id:int,
                                        provider_id:int, args: dict=Body(...),
                                        user: AuthUserModel = auth_spec()):
        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)
            return interest.generate_from_provider(provider_id, args=args, auth_user=user, session=session)

    async def set_sequence_default_duration(self, request:Request, id:int,
                                            duration:int = 10000,
                                            user: AuthUserModel = auth_spec()):
        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)
            return interest.sequence_set_default_duration(default_duration=duration, auth_user=user, session=session)

    async def get_sequence_contents(self, request: Request, id: int,
                                    full=False, user: AuthUserModel = auth_spec()):
        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)
            return interest.sequence_get_contents(full=full, auth_user=user, session=session)

    async def add_to_sequence(self, request:Request, id:int, item: SequenceAddTModel,
                              full=True, user: AuthUserModel = auth_spec()):
        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)
            result = interest.sequence_add(**item.dict(), auth_user=user, session=session)

        if not result:
            raise Exception

        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)
            return interest.sequence_get_contents(full=full, auth_user=user, session=session)

    async def remove_from_sequence(self, request:Request, id:int, position:int,
                                   full=True, user: AuthUserModel = auth_spec()):
        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)
            result = interest.sequence_remove(position=position, auth_user=user, session=session)

        if not result:
            raise Exception

        with get_session() as session:
            interest: MediaContentInterest = self._actual.item(id=id, session=session)
            return interest.sequence_get_contents(full=full, auth_user=user, session=session)

    async def change_item_duration(self, request:Request, id:int,
                                   position:int, duration:int,
                                   user: AuthUserModel = auth_spec()):
        pass

    def generate(self, name):
        desc = f'Content API for {titleize(singularize(name))} Interests'
        prefix = self._actual.interest_class.model.role_spec.prefix
        router = APIRouter(prefix=f'/{name}', tags=[desc],
                           dependencies=[Depends(authn_dependency)])

        router.add_api_route("/allowed_types", self.accepted_types, methods=["GET"],
                             response_model=Dict[str, ContentTypeDetailTModel],
                             response_model_exclude_none=True,
                             dependencies=[auth_spec(scopes=[f'{prefix}:read'])], )

        router.add_api_route("/{id}/content_info", self.content_info, methods=["GET"],
                             response_model=Union[MediaContentInfoFullTModel, MediaContentInfoTModel],
                             response_model_exclude_none=True,
                             dependencies=[auth_spec(scopes=[f'{prefix}:read'])])

        if 'media' in self._actual.accepted_types.keys():
            router.add_api_route("/{id}/formats/info/{format_id}", self.format_info, methods=["GET"],
                                 response_model=Union[MediaContentFormatInfoFullTModel, MediaContentFormatInfoTModel],
                                 response_model_exclude_none=True,
                                 dependencies=[auth_spec(scopes=[f'{prefix}:write'])])

            router.add_api_route("/{id}/formats/upload", self.upload_media_format, methods=["POST"],
                                 response_model=GenericTokenTModel,
                                 dependencies=[auth_spec(scopes=[f'{prefix}:write'])])

            # router.add_api_route("/{id}/formats/delete", self.delete_media_format, methods=["POST"],
            #                      # response_model=[],
            #                      dependencies=[auth_spec(scopes=[f'{prefix}:write'])])

        if 'structured' in self._actual.accepted_types.keys():
            router.add_api_route("/{id}/provider/{provider_id}/generate", self.generate_provider_content, methods=['POST'],
                                 # response_model=,
                                 dependencies=[auth_spec(scopes=[f'{prefix}:write'])])

        if 'sequence' in self._actual.accepted_types.keys():
            router.add_api_route("/{id}/sequence/contents", self.get_sequence_contents, methods=['GET'],
                                 dependencies=[auth_spec(scopes=[f'{prefix}:read'])])

            router.add_api_route("/{id}/sequence/duration", self.set_sequence_default_duration, methods=['POST'],
                                 response_model=SequenceDefaultDurationResponseTModel,
                                 dependencies=[auth_spec(scopes=[f'{prefix}:write'])])

            router.add_api_route("/{id}/sequence/add", self.add_to_sequence, methods=['POST'],
                                 # response_model=,
                                 dependencies=[auth_spec(scopes=[f'{prefix}:write'])])

            router.add_api_route("/{id}/sequence/remove/{position}", self.remove_from_sequence, methods=['POST'],
                                 # response_model=,
                                 dependencies=[auth_spec(scopes=[f'{prefix}:write'])])

        return [router]



# Copyright (C) 2019 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Media Content Configuration Options
====================================
"""


from tendril.utils.config import ConfigOption
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)

depends = ['tendril.config.core']

config_elements_media = [
    ConfigOption(
        'MEDIA_VIDEO_EXTENSIONS',
        "['.avi', '.mov', '.mp4', '.ogv', '.webm', '.wmv']",
        "List of recognized extensions for video files."
    ),
    ConfigOption(
        'MEDIA_IMAGE_EXTENSIONS',
        "['.jpg', '.png', '.gif']",           # WebP, SVG, Tiff don't play well with MediaInfo
        "List of recognized extensions for video files."
    ),
    ConfigOption(
        'MEDIA_DOCUMENT_EXTENSIONS',
        "['.pdf']",
        "List of recognized extensions for document files."
    ),
    ConfigOption(
        'MEDIA_EXTRA_EXTENSIONS',
        "[]",
        "List of extra extensions to recognize as media files. Note "
        "that this is only provided for highly specialized cases and "
        "for development-time use. Extensions listed here will "
        "probably break other code in unpredictable ways if used."
    ),
    ConfigOption(
        'MEDIA_EXTENSIONS',
        "MEDIA_VIDEO_EXTENSIONS + MEDIA_IMAGE_EXTENSIONS + MEDIA_DOCUMENT_EXTENSIONS + MEDIA_EXTRA_EXTENSIONS",
        "List of recognized extensions for media files"
    ),
    ConfigOption(
        'MEDIA_THUMBNAIL_SIZES',
        "[128,256,512]",
        "List of thumbnail sizes to generate for media content. If integers are provided, "
        "thumbnails will be square with MEDIA_THUMBNAIL_BACKGROUND in the remaining area. "
        "If a different shape is desired, provide tuples of two elements. Integers and "
        "tuples can be mixed."
    ),
    ConfigOption(
        'MEDIA_THUMBNAIL_BACKGROUND',
        "None",
        "If the thumbnail is to be paaded, then the background color to use for thumbnails. "
        "Leave as None to disable thumbnail padding. In principle, padding is "
        "relatively inexpensive and can provide well-sized images. However, if the "
        "background color uses an alpha channel, thumbnails will be generated in PNG and "
        "will be about 10 times larger, so use only if bandwidth is not a concern at all "
        "and latency is very low - essentially only for LAN deployments."
        "(255,255,255) creates a letterbox effect with a white background. Provide RGBA "
        "colors such as (255,255,255, 0) instead if transparency is needed. "
    ),
    ConfigOption(
        'MEDIA_UPLOAD_FILESTORE_BUCKET',
        '"incoming"',
        "The filestore bucket in which to write uploaded media files. Note that filestore "
        "will not have this bucket by default. You must create it or choose one that exists."
    ),
    ConfigOption(
        'MEDIA_PUBLISHING_FILESTORE_BUCKET',
        '"cdn"',
        "The filestore bucket in which published media files are to be written Note that "
        "filestore will not have this bucket by default. You must create it or choose one "
        "that exists."
    )
]


def load(manager):
    logger.debug("Loading {0}".format(__name__))
    manager.load_elements(config_elements_media,
                          doc="Media Content Configuration")



__path__ = __import__('pkgutil').extend_path(__path__, __name__)


from tendril.db.models.content import ContentModel


def _get_content_models():
    return {k: v.class_ for k, v in ContentModel.__mapper__.polymorphic_map.items()}


content_models = _get_content_models()


def _get_content_types():
    return set(content_models.keys())


content_types = _get_content_types()





from sqlalchemy.orm.exc import NoResultFound
from tendril.utils.db import with_db

from tendril.db.models.content import ContentModel
from tendril.db.models.content_formats import FileMediaContentFormatModel
from tendril.db.models.content_thumbnails import MediaContentFormatThumbnailModel
from tendril.db.models.content import SequenceContentAssociationModel
from tendril.db.controllers.interests import get_interest
from tendril.filestore.db.controller import get_stored_file
from tendril.structures.content import content_models


def _type_discriminator(type):
    if not type:
        return ContentModel
    if isinstance(type, str):
        return content_models[type]
    if issubclass(type, ContentModel):
        qmodel = type
    return qmodel


@with_db
def get_content(id=None, type=None, raise_if_none=True, session=None):
    filters = []
    qmodel = _type_discriminator(type)
    filters.append(qmodel.id == id)
    q = session.query(qmodel).filter(*filters)
    try:
        return q.one()
    except NoResultFound:
        if raise_if_none:
            raise
        return None


@with_db
def create_content(id=None, type=None, session=None, **kwargs):
    try:
        existing = get_content(id=id, session=session)
    except NoResultFound:
        pass
    else:
        raise ValueError(f"Could not create content container with "
                         f"ID {id}. Already Exists.")
    model = _type_discriminator(type)
    content = model(id=id)
    session.add(content)
    session.flush()
    return content


@with_db
def create_content_format_thumbnail(id=None, stored_file_id=None,
                                    width=None, height=None, session=None):
    try:
        thumbnail_instance = MediaContentFormatThumbnailModel(
            format_id=id,
            width=width,
            height=height,
            stored_file_id=stored_file_id
        )
        session.add(thumbnail_instance)
        session.flush()
        return thumbnail_instance
    except NoResultFound:
        raise ValueError(f"Could not find a content format "
                         f"with the provided id {id}")


@with_db
def create_content_format_file(id=None, stored_file_id=None,
                               width=None, height=None, duration=None,
                               info=None, session=None):
    try:
        _ = get_content(id=id, type='media', session=session)
        format_instance = FileMediaContentFormatModel(
            stored_file_id=stored_file_id,
            content_id=id,
            width=width,
            height=height,
            duration=duration,
            info=info
        )
        session.add(format_instance)
        session.flush()
        return format_instance
    except NoResultFound:
        raise ValueError(f"Could not find a {type} content file "
                         f"container with the provided id {id}")


@with_db
def get_format_by_stored_file(stored_file_id=None, raise_if_none=True, session=None):
    filters = []
    filters.append(FileMediaContentFormatModel.stored_file_id == stored_file_id)
    q = session.query(FileMediaContentFormatModel).filter(*filters)
    try:
        return q.one()
    except NoResultFound:
        if raise_if_none:
            raise
        return None


@with_db
def get_format_by_filename(filename=None, bucket=None, session=None):
    stored_file = get_stored_file(filename=filename, bucket=bucket, session=session)
    return get_format_by_stored_file(stored_file_id=stored_file.id, session=session)


@with_db
def get_content_by_filename(filename=None, content_type=None, bucket=None, session=None):
    format = get_format_by_filename(filename=filename, bucket=bucket, session=session)
    return get_content(id=format.content_id, type=content_type, session=session)


@with_db
def sequence_next_position(id=None, session=None):
    try:
        sequence = get_content(id=id, type='sequence', session=session)
        return max([x.position for x in sequence.contents], default=-1) + 1
    except NoResultFound:
        raise ValueError(f"Could not find a 'sequence' content "
                         f"container with the provided id {id}")


@with_db
def sequence_get_at_position(id, position, session=None):
    try:
        sequence = get_content(id=id, type='sequence', session=session)
        for x in sequence.contents:
            if x.position == position:
                return x
    except NoResultFound:
        raise ValueError(f"Could not find a 'sequence' content "
                         f"container with the provided id {id}")


@with_db
def sequence_prep_position(id, position, session=None):
    existing = sequence_get_at_position(id, position, session=session)
    if not existing:
        return
    sequence_prep_position(id, position + 1, session=session)
    existing.position = position + 1
    session.flush()

@with_db
def sequence_get_contents(id, session=None):
    try:
        sequence = get_content(id=id, type='sequence', session=session)
    except NoResultFound:
        raise ValueError(f"Could not find a 'sequence' content "
                         f"container with the provided id {id}")
    return [{
        'position': c.position,
        'duration': c.duration,
        'content': c.content,
    } for c in sequence.contents]


@with_db
def sequence_add_content(id, content, position=None, duration=None, session=None):
    content_id = content
    if position is None:
        position = sequence_next_position(id=id, session=session)
    else:
        sequence_prep_position(id, position, session=session)
    if not content_id:
        raise ValueError(f"Don't have a valid content_id. Got {content}")
    association = SequenceContentAssociationModel(sequence_id=id,
                                                  content_id=content_id,
                                                  position=position,
                                                  duration=duration)
    session.add(association)
    session.commit()


@with_db
def sequence_remove_content(id, position, session=None):
    assn = sequence_get_at_position(id=id, position=position, session=session)
    if not assn:
        raise ValueError(f"Sequence does not seem to have any "
                         f"content at position {position}.")
    session.delete(assn)
    session.commit()


@with_db
def sequence_pull_back_position(id, position, to_position, session=None):
    assn = sequence_get_at_position(id=id, position=position, session=session)
    if not assn:
        sequence_pull_back_position(id, position + 1, to_position, session=session)
    else:
        assn.position = to_position
        session.commit()


@with_db
def sequence_heal_positions(id=None, session=None):
    try:
        sequence = get_content(id=id, type='sequence', session=session)
        session.expire(sequence)
        if not len(sequence.contents):
            return
        for position in range(len(sequence.contents)):
            if not sequence_get_at_position(id=id, position=position, session=session):
                sequence_pull_back_position(id=id, position=position + 1, to_position=position, session=session)
    except NoResultFound:
        raise ValueError(f"Could not find a 'sequence' content "
                         f"container with the provided id {id}")



from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from .manager import ContentProviderManager
_manager = ContentProviderManager(prefix='tendril.structures.content.providers')

import sys
sys.modules[__name__] = _manager
_manager.finalize()



from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from .manager import InterestManager
_manager = InterestManager(prefix='tendril.interests')

import sys
sys.modules[__name__] = _manager
_manager.finalize()





from tendril.apiserver.templates.content import InterestContentRouterGenerator
from tendril.structures.content import content_models
from tendril.structures.content import content_types


class ContentLibraryMixin(object):
    media_types_allowed = content_types
    _additional_api_generators = [InterestContentRouterGenerator]

    def __init__(self, *args, **kwargs):
        super(ContentLibraryMixin, self).__init__(*args, **kwargs)
        self._accepted_types = None

    @property
    def accepted_types(self):
        if not self._accepted_types:
            library_allowed_types = self.media_types_allowed
            if library_allowed_types:
                accepted_types = content_types.intersection(library_allowed_types)
            else:
                accepted_types = content_types
            self._accepted_types = {k: v
                                    for (k, v) in content_models.items()
                                    if k in accepted_types}
        return self._accepted_types
